//
//  agent.c
//  Infektsimulation
//
//  Created by //// on 20.08.13.
//  Copyright (c) 2013 Alexander Droste & Florian Flachsenberg. All rights reserved.
//


#include "agent.h"
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <gsl_rng.h>
#include <assert.h>
#include <string.h>

#define INVALID_AGENT_TYPE_ERR 666
//global number generator
extern gsl_rng* rng;
//later one rng must be created for each process,
//to avoid communication between processes
//http://stackoverflow.com/questions/8770019/creating-random-numbers-in-a-multithreaded-program-in-c


//–––––––––––––––––––private functions––––––––––––––––––
static unsigned long agent_create_immunogenicity_factor();
//–––––––––––––––––––––––––––––––––––––––––––––––––––––

//TODO: klären wie lang ein Zeitschritt sein soll
//-> Annahme eine Sekunde


/*no need to build agent_new_individual-function explicitly 
 must be called with calloc, new_agent->xy=..., by hand anyway*/

bool agent_is_valid (Agent* agent)
{
    if (agent->mortality < 0.0 || agent->mortality > 100.0) return false;
    if (agent->immunogenicity < 0.0 || agent->immunogenicity > 100.0) return false;
            
    if (agent->infection_prob.air < 0.0 || agent->infection_prob.air > 100.0) return false;
    if (agent->infection_prob.skin < 0.0 || agent->infection_prob.skin > 100.0) return false;
    if (agent->infection_prob.oral < 0.0 || agent->infection_prob.oral > 100.0) return false;
    if (agent->infection_prob.xxx < 0.0 || agent->infection_prob.xxx > 100.0) return false;

    if (agent->outbreak_prob < 0.0 || agent->outbreak_prob > 100.0) return false;
    if (agent->mutation_prob < 0.0 || agent->mutation_prob > 100.0) return false;
            
    /* difference between mean and max time may not be larger than mean time */
    if (2* agent->incubation_mean_dur <= agent->incubation_max_dur) return false;
    if (2* agent->illness_mean_dur <= agent->illness_max_dur) return false;
 
    return true;
}

Agent* agent_new_stereotype(AgentType agent_type)
{
    Agent* new_agent = calloc(1, sizeof(Agent));
    new_agent->mutation_counter = 0;

    //TODO: create "real" agent types
    switch (agent_type) {
        case Influenza:
            
            new_agent->mortality = 0.20;
            new_agent->immunogenicity = 5.00;
            
            new_agent->infection_prob.air = 5.00;
            new_agent->infection_prob.skin = 0.00;
            new_agent->infection_prob.oral = 20.00;
            new_agent->infection_prob.xxx = 0.00;
            new_agent->outbreak_prob = 20.00;
            new_agent->mutation_prob = 10.00;

            
            new_agent->incubation_mean_dur = 4;
            new_agent->incubation_max_dur = 6;
            new_agent->illness_mean_dur = 6;
            new_agent->illness_max_dur = 10;
            new_agent->infectious_incubation = false;
            
            new_agent->immunogenicity_factor = 1654L;
            new_agent->mutation_counter_threshold = 2;
            break;

            //another common disease
        case Smallpox:
            new_agent->mortality = 32.50;
            new_agent->immunogenicity = 95.00;
            
            new_agent->infection_prob.air = 70.00;
            new_agent->infection_prob.skin = 40.00;
            new_agent->infection_prob.oral = 90.00;
            new_agent->infection_prob.xxx = 4.00;
            new_agent->outbreak_prob = 90.00;
            //new_agent->mutation_prob = 30.00;
            new_agent->mutation_prob = 1.00;
            
            new_agent->incubation_mean_dur = 3;
            new_agent->incubation_max_dur = 5;
            new_agent->illness_mean_dur = 20;
            new_agent->illness_max_dur = 30;
            new_agent->infectious_incubation = 0;

            new_agent->immunogenicity_factor = 10001L;
            new_agent->mutation_counter_threshold = 2;
            break;

           
        default:
            //if no valid agent_type was chosen
            //->should not happen
            printf("invalid agent type chosen");
            exit(INVALID_AGENT_TYPE_ERR);
            
            break;
    }
    return new_agent;
}

Agent* agent_copy(Agent* inc_agent)
{
    Agent* new_agent = calloc(1, sizeof(Agent));
    memcpy(new_agent, inc_agent, sizeof(Agent));
    //do manual allocation and assignment if there are pointer types in struct
    return new_agent;
}

void agent_delete(Agent* inc_agent)
{
    free(inc_agent);
    inc_agent = NULL;
}
                   
void agent_print_properties(Agent* inc_agent)
{
    printf("agent properties\n\n                                \
                                                                \
           mortality: %f\n                                      \
           immunogenicity %f\n                                  \
           infection_prob air %f oral%f, skin %f, xxx %f\n      \
           outbreak_prob %f\n                                   \
           mutation_prob %f\n                                   \
           incubation_mean_dur %d\n                             \
           incubation_max_dur %d\n                              \
           illness_mean_dur %d\n                                \
           illness_max_dur %d\n                                 \
           infectious_incubation %i\n,                          \
           immunogenicity_factor %lu\n",
           
           inc_agent->mortality,
           inc_agent->immunogenicity,
           inc_agent->infection_prob.air,
           inc_agent->infection_prob.oral,
           inc_agent->infection_prob.skin,
           inc_agent->infection_prob.xxx,
           inc_agent->outbreak_prob,
           inc_agent->mutation_prob,
           inc_agent->incubation_mean_dur,
           inc_agent->incubation_max_dur,
           inc_agent->illness_mean_dur,
           inc_agent->illness_max_dur,
           inc_agent->infectious_incubation,
           inc_agent->immunogenicity_factor);
}

void agent_mutate (Agent* agent)
{
    /* add mutation of values */
    agent->mutation_counter++;

    /* if mutation threshold value is reached => set to 0 and change
       immunogenicity */
    if (agent->mutation_counter == agent->mutation_counter_threshold) {
        agent->mutation_counter = 0;
        agent->immunogenicity_factor = agent_create_immunogenicity_factor();
    }
       
    enum {mortality, immunogenicity, outbreak_prob, mutation_prob};
    very_short_int property_number = rnd_int(rng)%4;
       
    //mutate one property by max 1%
    switch (property_number) {
        case mortality:
            agent->mortality = keep_in_range(0, 100,
                               rnd_sign(rng)*rnd_one_percent(rng) + agent->mortality);
            break;
        case immunogenicity:
            agent->immunogenicity = keep_in_range(0, 100,
                                    rnd_sign(rng)*rnd_one_percent(rng) + agent->immunogenicity);
            break;
        case outbreak_prob:
            agent->outbreak_prob = keep_in_range(0, 100,
                                   rnd_sign(rng)*rnd_one_percent(rng) + agent->outbreak_prob);
            break;
        case mutation_prob:
            agent->mutation_prob = keep_in_range(0, 100,
                                   rnd_sign(rng)*rnd_one_percent(rng) + agent->mutation_prob);
            break;

    }
}

/**
 *  Create a new agent identifier
 *  Use only in case of mutation.
 *
 *  @return identifier
 */
static unsigned long agent_create_immunogenicity_factor()
{
    return ((unsigned long)(rnd_int(rng) % USHRT_MAX));
}

