//
// host.c
// Infektsimulation
//
// Created by //// on 20.08.13.
// Copyright (c) 2013 Alexander Droste & Florian Flachsenberg. All rights reserved.
//


#include <stdlib.h>
#include <stdio.h>

#include "agent.h"
#include "host.h"
#include "datatypes.h"
#include "world.h"
#include <gsl_rng.h>
#include <gsl_randist.h>
#include <limits.h>

//accessing global vars with extern
extern World world;
extern gsl_rng *rng;

static void host_find_initial_pos(Host *host);

//get random values - man pages
//https://www.gnu.org/software/gsl/manual/html_node/Sampling-from-a-random-number-generator.html


Host *host_create()
{
	Host *new_host = calloc(1, sizeof(Host));
	new_host->state = HEALTHY;
	new_host->agent = NULL;
	new_host->field = NULL;
	//max value of random function is half the size of unsigned int max
    
	//find free position for host
    host_find_initial_pos(new_host);
	//minimum of 24 + random[0, 48]
	new_host->disappear_duration = 24 + (rnd_int(rng) % 49);
    
	//immunity list
	new_host->immunities = 0L;
	return new_host;
}

void host_delete(Host *host)
{
	//agent initialised with NULL->calling free(NULL) is ok
	free(host->agent);
    
	/* list_delete_element is called in world_remove_host_from_field */
	free(host);
	host = NULL;
}

void host_delete_agent(Host *host)
{
	agent_delete(host->agent);
	host->agent = NULL;
}

/**
 *  Find a position for a host during creation process.
 *  Additional function to prevent the possibility of a deadlock
 *  which could happen in host_move during creation process. 
 *  If a host is placed on a reserved field which is surrounded 
 *  by reserved fields there's no way to find a free position.
 *  If not all positions in world are reserved, host_find_initial_pos
 *  will find a free one. Despite, this removes the check for 
 *  host->field != NULL in host_move.
 *
 *  @param host host to find position for
 */
static void host_find_initial_pos(Host *host)
{
	host->pos.x = rnd_int(rng) % world.size.width; //pos x/y is unsigned int
	host->pos.y = rnd_int(rng) % world.size.height;

	bool keep_searching = true;    
	while (keep_searching) {        
		host->pos.x = (rnd_int(rng) % world.size.width);
		host->pos.y = (rnd_int(rng) % world.size.height);
        
		//if another host is on the same position, keep searching
		keep_searching = world_host_position_collides(host);
	}
}

void host_move(Host *host)
{
	bool keep_searching = true;
	unsigned old_x_pos = host->pos.x, old_y_pos = host->pos.y;
	very_short_int rnd_dir_x, rnd_dir_y;
    
	while (keep_searching) {
		//possible values rand%3={0,1,2} -> -1 ={-1, 0, 1}
		rnd_dir_x = (rnd_int(rng) % 3) - 1;
		rnd_dir_y = (rnd_int(rng) % 3) - 1;
        
		host->pos.x = (old_x_pos + rnd_dir_x + world.size.width) % world.size.width;
		host->pos.y = (old_y_pos + rnd_dir_y + world.size.height) % world.size.height;
        
		//if another host is on the same position, keep searching
		keep_searching = world_host_position_collides(host);
	}
    
	/* If the host was assigned to a field before
     and should be moved due to his new position. */
	Field *correct_field = world_get_field_for_host(host);
	if (host->field != correct_field) {
		world_transfer_host_to_field(host, correct_field);
	}
}

void host_start_disease(Host *host)
{
	host->state = ILL;
	world.ill_total++;
  world.ill_current++;
  world.infected_current--;
  int illness_time = host->end_of_disease - host->outbreak_time;
  if (illness_time > world.max_illness) {
      world.max_illness = illness_time;
  }
  else if (illness_time < world.min_illness) {
      world.min_illness = illness_time;
  }

  world.mean_illness = world.mean_illness + 
    (double)(illness_time - world.mean_illness)/world.ill_total;


}

void host_end_disease(Host *host)
{
  if (host->state == INFECTED) {
    world.infected_current--;
  }
  if (host->state == ILL) {
    world.ill_current--;
  }
	if (host->will_die) {
		host->state = DEAD;
		world.died_total++;
    world.infectious_dead_current++;
	}
	else if (host->will_immune) {
		host->state = IMMUNE;
    host->immunities |= host->agent->immunogenicity_factor;
		world.immune_total++;
	}
	else {
		host->state = HEALTHY;
	}
	if (host->state == IMMUNE || host->state == HEALTHY) {
		list_transfer_element(host->field->healthy, host->field->infected, host->list_element);
		host_delete_agent(host);
    world.healthy_current++;
    host->field->nof_infected--;
	}
}


void host_process(Host *host, unsigned short inc_time) {
	if (host->state == DEAD) {
		//after hosts die they stay for the time of their disappear_duration
		host->disappear_duration -= 1;
		//if the duration is over they get removed completely
		if (host->disappear_duration == 0) {
      host->field->nof_infected--;
			world_remove_host_from_field(host);
			host_delete(host);
      world.infectious_dead_current--;
		}
	}
    /* if not dead */
	else {
        /* process infection if host is not healthy */
		if (host->state != HEALTHY && host->state != IMMUNE) {
            /* at outbreak time host infection will turn into disease or not */
			if (inc_time == host->outbreak_time && host->will_ill) {
                /* start disease or skip it*/
                host_start_disease(host);
			}
            /* if disease time is over => end disease */
			else if (inc_time == host->end_of_disease) {
                host_end_disease(host);
			}
		}
	}
}

/**
 * In case of a infection this function is called.
 *
 * @param host host to get infected
 * @param inc_agent agent infecting the host
 * @param inc_time time point to capture start of infection
 */
void host_infect(Host *host, Agent *inc_agent, unsigned inc_time)
{
  bool is_contained = ((host->immunities & inc_agent->immunogenicity_factor)
      == inc_agent->immunogenicity_factor);

  if (is_contained) {return;}
	/* copy agent from other host */
	host->agent = agent_copy(inc_agent);
	host->infection_time = inc_time;
	host->state = INFECTED;
  world.healthy_current--;
	list_transfer_element(host->field->infected, host->field->healthy, host->list_element);
	world.infected_total++;
  world.infected_current++;
  host->field->nof_infected++;
    
	if (rnd_percent(rng) < host->agent->mutation_prob) {
		agent_mutate(host->agent);
	}
    
	//immunogenicity checked against random value
	if (rnd_percent(rng) < inc_agent->immunogenicity) {
		host->will_immune = 1;
	}
	else {
		host->will_immune = 0;
	}
    
	//outbreak time gets calculated
	int diff = inc_agent->incubation_max_dur - inc_agent->incubation_mean_dur;
	int time_variation = (rnd_int(rng) % (2 * diff + 1)) - diff;
  int incubation_time = inc_agent->incubation_mean_dur + time_variation;
	host->outbreak_time = inc_time + incubation_time;
  if (incubation_time > world.max_incubation) {
      world.max_incubation = incubation_time;
  }
  else if (incubation_time < world.min_incubation) {
      world.min_incubation = incubation_time;
  }

  world.mean_incubation = world.mean_incubation + 
    (double)(incubation_time - world.mean_incubation)/world.infected_total;


	//outbreak probability checked against random value
	if (rnd_percent(rng) < inc_agent->outbreak_prob) {
		host->will_ill = 1;
        
		//mortality is checked against random value
		if (rnd_percent(rng) < inc_agent->mortality) {
			host->will_die = 1;
		}
		else {
			host->will_die = 0;
		}
        
		//disease endpoint gets calculated        
		diff = inc_agent->illness_max_dur - inc_agent->illness_mean_dur;
		int time_variation_2 = (rnd_int(rng) % (2 * diff + 1)) - diff;
        //disease time calculated beginning with outbreak_time!!
		host->end_of_disease = host->outbreak_time + inc_agent->illness_mean_dur + time_variation_2;

	}
	else {
		host->will_ill = 0;
		host->will_die = 0;
        host->end_of_disease = host->outbreak_time;
	}
}


static void host_infection_process_for_field(Host* host_to_infect, Field *field, unsigned short time) {
	List_element *el = list_first(field->infected);
	Host* neigh, *other_host_in_range;

  Host *neigh_array[field->nof_infected];
  int j = 0;
	while (el) {
		neigh = list_element_get_value(el);
    el = list_element_next(el);
    neigh_array[j] = neigh;
    ++j;
  }

  if (field->nof_infected > 1) {
      gsl_ran_shuffle(rng, neigh_array, field->nof_infected, sizeof(Host*));
  }

  int i;
  for (i = 0; i < field->nof_infected; ++i)
  { 
		neigh = neigh_array[i];
		// list contains only infected and ill host neighbours
		// => are infectious if ill or agent is infectious during incubation
		if ((neigh->agent->infectious_incubation && neigh->state == INFECTED && time > neigh->infection_time) ||
		    (neigh->state == ILL && ((!neigh->agent->infectious_incubation && time > neigh->outbreak_time) ||
              (neigh->agent->infectious_incubation && time > neigh->infection_time))) || neigh->state == DEAD)
              {
                  unsigned dist;
                  other_host_in_range = neigh;
                  dist = sq_euclidean(host_to_infect->pos.x, host_to_infect->pos.y,
                                      other_host_in_range->pos.x, other_host_in_range->pos.y);
                  
                  if (dist < world.radius_air_sq && other_host_in_range->state != DEAD) {
                      if (rnd_percent(rng) < other_host_in_range->agent->infection_prob.air) {
                          host_infect(host_to_infect, other_host_in_range->agent, time);
                          if (host_to_infect->state == INFECTED) {break;}
                      }
                  }
                  if (dist < world.radius_oral_sq && other_host_in_range->state != DEAD) {
                      if (rnd_percent(rng) < other_host_in_range->agent->infection_prob.oral) {
                          host_infect(host_to_infect, other_host_in_range->agent, time);
                          if (host_to_infect->state == INFECTED) {break;}
                      }
                  }
                  if (dist < world.radius_skin_sq) {
                      if (rnd_percent(rng) < other_host_in_range->agent->infection_prob.skin) {
                          host_infect(host_to_infect, other_host_in_range->agent, time);
                          if (host_to_infect->state == INFECTED) {break;}
                      }
                  }
                  if (dist < world.radius_xxx_sq && other_host_in_range->state != DEAD) {
                      if (rnd_percent(rng) < other_host_in_range->agent->infection_prob.xxx) {
                          host_infect(host_to_infect, other_host_in_range->agent, time);
                          if (host_to_infect->state == INFECTED) {break;}
                      }
                  }
              }
//		el = list_element_next(el);
	}
}

/**
 * Interacts with all other host in range.
 *
 *
 * @param host host to interact with others
 * @param time time-point of interaction
 */
void host_get_infection(Host *host, unsigned short time)
{
	Coordinates field_coord = world_get_field_coord_for_host(host);
	Field *wf = world_get_field_for_host(host);
	Position hostPos = world_get_pos_within_field(host->pos.x, host->pos.y);
    
	unsigned field_max_x_pos = world.size_of_one_field.width - 1;
	unsigned field_max_y_pos = world.size_of_one_field.height - 1;
    
	//squared euclidean distances to each corner
	unsigned left = sq_euclidean(0, hostPos.y, hostPos.x, hostPos.y);
	unsigned topleft = sq_euclidean(0, 0, hostPos.x, hostPos.y);
	unsigned top = sq_euclidean(hostPos.x, 0, hostPos.x, hostPos.y);
	unsigned topright = sq_euclidean(field_max_x_pos, 0, hostPos.x, hostPos.y);
	unsigned right = sq_euclidean(field_max_x_pos, hostPos.y, hostPos.x, hostPos.y);
	unsigned bottomright = sq_euclidean(field_max_x_pos, field_max_y_pos, hostPos.x, hostPos.y);
	unsigned bottom = sq_euclidean(hostPos.x, field_max_y_pos, hostPos.x, hostPos.y);
	unsigned bottomleft = sq_euclidean(0, field_max_y_pos, hostPos.x, hostPos.y);
  
    Field* neighbourfields[9];
    unsigned fields = 0;
    unsigned i;
    
	unsigned matrix_cols = world.field_matrix->n;
	unsigned matrix_rows = world.field_matrix->m;

    neighbourfields[fields++] = wf;  
    
	if (left < world.maxradius_sq) {
		wf = get_field_from_matrix_for_coord((field_coord.x - 1 + matrix_cols) % matrix_cols, field_coord.y);
        neighbourfields[fields++] = wf;
	}
	if (topleft < world.maxradius_sq) {
		wf = get_field_from_matrix_for_coord((field_coord.x - 1 + matrix_cols) % matrix_cols,
                                             (field_coord.y - 1 + matrix_rows) % matrix_rows);
        neighbourfields[fields++] = wf;
	}
	if (top < world.maxradius_sq) {
		wf = get_field_from_matrix_for_coord(field_coord.x, (field_coord.y - 1 + matrix_rows) % matrix_rows);
        neighbourfields[fields++] = wf;
	}
	if (topright < world.maxradius_sq) {
		wf = get_field_from_matrix_for_coord((field_coord.x + 1 + matrix_cols) % matrix_cols,
                                             (field_coord.y - 1 + matrix_rows) % matrix_rows);
        neighbourfields[fields++] = wf;
	}
	if (right < world.maxradius_sq) {
		wf = get_field_from_matrix_for_coord((field_coord.x + 1 + matrix_cols) % matrix_cols, field_coord.y);
        neighbourfields[fields++] = wf;
	}
	if (bottomright < world.maxradius_sq) {
		wf = get_field_from_matrix_for_coord((field_coord.x + 1 + matrix_cols) % matrix_cols,
                                             (field_coord.y + 1 + matrix_rows) % matrix_rows);
        neighbourfields[fields++] = wf;
	}
	if (bottom < world.maxradius_sq) {
		wf = get_field_from_matrix_for_coord(field_coord.x, (field_coord.y + 1 + matrix_rows) % matrix_rows);
        neighbourfields[fields++] = wf;
	}
	if (bottomleft < world.maxradius_sq) {
		wf = get_field_from_matrix_for_coord((field_coord.x - 1 + matrix_cols) % matrix_cols,
                                             (field_coord.y + 1 + matrix_rows) % matrix_rows);
        neighbourfields[fields++] = wf;
	}

    gsl_ran_shuffle(rng, neighbourfields, fields, sizeof(*neighbourfields));

    for (i = 0; i < fields; i++) {
        host_infection_process_for_field(host, neighbourfields[i], time);
        if (host->state == INFECTED) return;
    }
}
