//
//  main.c
//  Infektsimulation
//
//  Created by //// on 20.08.13.
//  Copyright (c) 2013 Alexander Droste & Florian Flachsenberg. All rights reserved.
//

#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include "host.h"
#include "agent.h"
#include "world.h"
#include <gsl_rng.h> //gsl - high quality random number generators

gsl_rng* rng; //rng = random number generator

extern World world;

int main(int argc, char** argv)
{

    if (argc < 6) {
        fprintf(stderr,"usage: %s world_size number_hosts percentage_infected\n"\
                "number_of_steps agent stop-flag\nworld_size: length of world axis\n"\
                "number_hosts: initial number of hosts\n"\
                "percenta_infected: percentage of hosts initially infected\n"\
                "number_of_steps: number of time steps of simulation\n"\
                "agent: agenty type ( )\n"\
                "stop-flag: 0, if no stopping if all infectious hosts are gone\n", argv[0]);
        return 1;
    }

    unsigned world_size, number_hosts;
    float percentage_infected;
    unsigned time_steps, agent_code, stop_if_finished;

    if (sscanf(argv[1], "%u", &world_size) != 1) {
        fprintf(stderr,"error: %s is no valid world size!\n", argv[1]);
        return 1;
    }

    if (sscanf(argv[2], "%u", &number_hosts) != 1) {
        fprintf(stderr,"error: %s is no valid number of hosts!\n", argv[2]);
        return 1;
    }

    if (sscanf(argv[3], "%f", &percentage_infected) != 1 || 
        percentage_infected < 0.0 || percentage_infected > 100.0) {
        fprintf(stderr,"error: %s is no valid infection percentage!\n", argv[3]);
        return 1;
    }

    if (sscanf(argv[4], "%u", &time_steps) != 1) {
        fprintf(stderr,"error: %s is no valid number of time steps!\n", argv[4]);
        return 1;
    }

    if (sscanf(argv[5], "%u", &agent_code) != 1 || agent_code > 5) {
        fprintf(stderr,"error: %s is no valid agent code!\n", argv[5]);
        return 1;
    }

    if (sscanf(argv[6], "%u", &stop_if_finished) != 1 ||
        stop_if_finished > 1) {
        fprintf(stderr,"error: %s is no valid mode for stop-flag (0 or 1)!\n", argv[6]);
        return 1;
    }

    rng = gsl_rng_alloc(gsl_rng_knuthran);
    gsl_rng_set(rng, 0);
    //http://fossies.org/dox/gsl-1.16/knuthran_8c_source.html

    unsigned i;
    Agent* new_agent = agent_new_stereotype(agent_code);
    if (!agent_is_valid(new_agent)) {
        fprintf(stderr, "specified agent is not valid!\n");
        return 1;
    }
    
    //create world with size (axes length^2), agent
    world_create_for_agent(world_size, new_agent);
    //create i hosts and add them to the world
    for (i = 0; i < number_hosts; ++i) {
        Host* new_host = host_create();
        world_add_host_to_field(new_host);
        //printf("%u, %u\n", new_host->pos.x, new_host->pos.y);
        world.healthy_current++;
        if (i < ceil(number_hosts * (percentage_infected/100.0))) {
            host_infect(new_host, new_agent, 0);
        }
    }
    
    // simulate x time steps
    world_simulate(time_steps, stop_if_finished);

    printf("back calculated values (CAUTION, only valid without mutations)\n");
    printf("total infected: %u\n", world.infected_total);
    /* total ill: total infected times outbreak prob */
    printf("total ill: %u (expected from total infected and agent outbreak prob: %.1f)\n", 
        world.ill_total, world.infected_total*new_agent->outbreak_prob/100.0);
    /* total died: total ill times mortality */
    printf("total died: %u (expected from total ill and agent mortality: %.1f)\n", 
        world.died_total, world.ill_total*new_agent->mortality/100.0);
    /* total immune: a little bit tricky; composed of infected and not ill
       (times immunogenicty) and ill and NOT dead (because death processed
       before immunity)
    */
    printf("total immune: %u (expected from total infected and immunogenic prob: %.1f)\n", 
        world.immune_total, 
        world.infected_total*new_agent->immunogenicity/100.0*(100.-new_agent->outbreak_prob)/100.0
        + world.ill_total*new_agent->immunogenicity/100.0*(100.-new_agent->mortality)/100.0);
    printf("incubation time (mean, min, max): %.1f, %u, %u (expected: %u, %u, %u)\n",
        world.mean_incubation, world.min_incubation, world.max_incubation,
        new_agent->incubation_mean_dur, 
        2*new_agent->incubation_mean_dur - new_agent->incubation_max_dur,
        new_agent->incubation_max_dur);
    printf("illness time (mean, min, max): %.1f, %u, %u (expected: %u, %u, %u)\n",
        world.mean_illness, world.min_illness, world.max_illness,
        new_agent->illness_mean_dur, 
        2*new_agent->illness_mean_dur - new_agent->illness_max_dur, 
        new_agent->illness_max_dur);

    return 0;
}
