//
//  world.c
//  Infektsimulation
//
//  Created by //// on 21.08.13.
//  Copyright (c) 2013 Alexander Droste & Florian Flachsenberg. All rights reserved.
//

#include "world.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <string.h>

#ifdef DRAW
#define MAGICKCORE_QUANTUM_DEPTH 16
#define MAGICKCORE_HDRI_ENABLE 0
#include <magick_wand.h>
#endif

//global world var can get accessed through the "extern" modifier in other files
World world;

//––––––––––––––––private functions––––––––––––––––––––––––––––––
static Coordinates world_get_field_coord_for_pos(unsigned int x, unsigned int y);
//–––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––


/**
 *  initialises, creates the world, defines world size, fields size
 *
 *  @param world_axes_length defines width and height of the world
 *  @param field_axes_length defines width and height for all fields
 */
void world_create(unsigned int world_axes_length, unsigned int field_axes_length)
{
    //use of static to construct singleton
    static bool  world_is_created = false;
    if (!world_is_created) {
        world.size.width = world_axes_length;  
        world.size.height = world_axes_length;
        
        world.size_of_one_field.width = field_axes_length;
        world.size_of_one_field.height = field_axes_length;

       
        //to ensure fields are created even if they are not used completely
        //example: world_axes_length = 100, field_axes_length = 12
        //100/12 = 8.3 -> 9 rows/cols are needed
        unsigned int rows = ceil(world_axes_length/field_axes_length);
        unsigned int cols = ceil(world_axes_length/field_axes_length);
        
        world.field_matrix = createMatrix(rows, cols);
        world_is_created = true;

        world.immune_total = 0;
        world.infected_total = 0;
        world.died_total = 0;
        world.ill_total = 0;
        world.ill_current = 0;
        world.infected_current = 0;
        world.healthy_current = 0;
        world.infectious_dead_current = 0;

        world.radius_air_sq = radius_air*radius_air;
        world.radius_skin_sq = radius_skin*radius_skin;
        world.radius_oral_sq = radius_oral*radius_oral;
        world.radius_xxx_sq = radius_xxx*radius_xxx;

        if (world.radius_air_sq > world.maxradius_sq) world.maxradius_sq = world.radius_air_sq;
        if (world.radius_oral_sq > world.maxradius_sq) world.maxradius_sq = world.radius_oral_sq;
        if (world.radius_skin_sq > world.maxradius_sq) world.maxradius_sq = world.radius_skin_sq;
        if (world.radius_xxx_sq > world.maxradius_sq) world.maxradius_sq = world.radius_xxx_sq;

        world.min_incubation = world.min_illness = USHRT_MAX;
        world.max_incubation = world.max_illness = 0;
        world.mean_incubation = world.mean_illness = 0.0;
    }
}

/**
 *  initialises, creates the world, defines world size, fields size
 *  is optimized for agent
 *  this function optimizes the field size and ignores zero-probabilty-transmissions
 *
 *  @param world_axes_length defines width and height of the world
 *  @param agent specifies Agent the parameters should be optimized for
 *
 */
void world_create_for_agent(unsigned world_axes_length, Agent* agent)
{
    unsigned maxr = 0;
    //find max radius of all transmission types
    if (agent->infection_prob.air != 0 && radius_air > maxr) maxr = radius_air;
    if (agent->infection_prob.skin != 0 && radius_skin > maxr) maxr = radius_skin;
    if (agent->infection_prob.oral != 0 && radius_oral > maxr) maxr = radius_oral;
    if (agent->infection_prob.xxx != 0 && radius_xxx > maxr) maxr = radius_xxx;

    world_create(world_axes_length, maxr*2);
    world.maxradius_sq = maxr*maxr;
}

/**
 *  Transfers a host to a another field
 *
 *  @param inc_host host to transfer
 *  @param dest_field field to transfer host to
 */
void world_transfer_host_to_field(Host* inc_host, Field* dest_field)
{
    #ifdef DEBUG
        assert(inc_host->field != NULL);
        assert(dest_field != NULL);
        assert(inc_host->field != dest_field);
    #endif
    
    List* dest_list;
    List* src_list;
    

    if(inc_host->state == HEALTHY || inc_host->state == IMMUNE){
        src_list = inc_host->field->healthy;
        dest_list = dest_field->healthy;
    }
    else{
        src_list = inc_host->field->infected;
        dest_list = dest_field->infected;
        inc_host->field->nof_infected--;
        dest_field->nof_infected++;
    }
    //destination field must be assigned to hosts attribute
    inc_host->field = dest_field;
    list_transfer_element(dest_list, src_list, inc_host->list_element);
}

/**
 *  Calculate the appropriate field for a host and add it to it
 *
 *  @param inc_host host to add to a field based on his position
 */
void world_add_host_to_field(Host* inc_host)
{
    List* list;
    Field* wf = world_get_field_for_host(inc_host);
    
    //host knows his field
    inc_host->field = wf;
    
    if (inc_host->state == HEALTHY || inc_host->state == IMMUNE) {
        list = wf->healthy;
    }
    else {
        list = wf->infected;
    }
    /* simplifies deletion -> list_delete_element
       used in world_remove_host_from_field */
    inc_host->list_element = list_append(list, inc_host);
}

/**
 *  Remove a host from his field.
 *  Should only be called after host died and
 *  dissappear duration has passed. Must be called before host_delete.
 *
 *  @param inc_host host to remove
 */
void world_remove_host_from_field(Host* inc_host)
{
    List* list;
    //the hosts assigned field must be used
    Field* wf = inc_host->field;
    if (inc_host->state == HEALTHY || inc_host->state == IMMUNE) {
        list = wf->healthy;
    }
    else {
      list = wf->infected;
    }
    inc_host->field = NULL;
    list_delete_element(list, inc_host->list_element);
}

/**
 *  Find out if another host is on the same position
 *
 *  @param inc_host the host to check the collision for
 *
 *  @return boolean value if there is a collision
 */
bool world_host_position_collides(Host* inc_host)
{

    Field* wf = world_get_field_for_host(inc_host);
    List_element* el_healthy = list_first(wf->healthy);
    List_element* el_infected = list_first(wf->infected);
    #ifdef PRINT_COLLISION
        static unsigned int colli = 0;
    #endif
    
    //initialised with collision set to 0
    bool collision = 0;

    //iterate over all elements in list
    while (el_healthy) {
        Host* host_from_list = el_healthy->data;

        //if the list element is not the given host itself
        if (host_from_list != inc_host) {
            if(host_from_list->pos.x == inc_host->pos.x &&
               host_from_list->pos.y == inc_host->pos.y){
                #ifdef PRINT_COLLISION
                    printf("collision no: %u\n", ++colli);
                #endif
                collision = 1; //if positions are equal
                break; //end while-loop
            }
        }
        el_healthy = list_element_next(el_healthy);

    }
    //if no collision in first loop
    if (!collision) {
        while (el_infected) {
            Host* host_from_list = el_infected->data;
            
            //if the list element is not the given host itself
            if (host_from_list != inc_host) {
                if(host_from_list->pos.x == inc_host->pos.x &&
                   host_from_list->pos.y == inc_host->pos.y){
                    #ifdef PRINT_COLLISION
                        printf("collision no: %u\n", ++colli);
                    #endif
                    collision = 1; //if positions are equal
                    break; //end while-loop
                }
            }
            el_infected = list_element_next(el_infected);
        }
    }
  
    return collision;
}

/**
 *  Get the correct field for a host, based on his position
 *  Which field is chosen is calculated in world_get_field_coord_for_pos
 *
 *  @param inc_host given host
 *
 *  @return field (list of hosts in same field)
 */
Field* world_get_field_for_host(Host* inc_host)
{
    Coordinates field_coordinates =
    world_get_field_coord_for_pos(inc_host->pos.x, inc_host->pos.y);
    
    Field* wf =
    get_field_from_matrix_for_coord(field_coordinates.x, field_coordinates.y);
    
    return wf;

}

/**
 *  Get field coordinates in matrix for position
 *
 *  @param x x-postion
 *  @param y y-postion
 *
 *  @return field coordinates
 *  
 */
static Coordinates world_get_field_coord_for_pos(unsigned int x, unsigned int y)
{
    Coordinates field_coordinates;
    #if DEBUG
        assert(!(x >= world.size.width || y >= world.size.height));
    #endif

    /*coordinates and width/height are all ints
     -> so division behaves as: floor(x/world.size_of_one_field.width);*/
    field_coordinates.x = x/world.size_of_one_field.width;
    field_coordinates.y = y/world.size_of_one_field.height;
    
    return field_coordinates;
}

/**
 *  Get the coordinates for the field the host is positioned in.
 *
 *  @param inc_host host to find out the field coordinates for
 *
 *  @return field coordinates
 */
Coordinates world_get_field_coord_for_host(Host* inc_host)
{
    return world_get_field_coord_for_pos(inc_host->pos.x, inc_host->pos.y);
}

/**
 *  Get the coordinates within a field for a "world"-position.
 *  The coordinates are in range of the general field width/height.
 *
 *  @param x x-position
 *  @param y y-position
 *
 *  @return coordinates within the field
 */
Position world_get_pos_within_field(unsigned x, unsigned y)
{
    Position coord;
    coord.x = x % world.size_of_one_field.width;
    coord.y = y % world.size_of_one_field.height;

    return coord;
}


#ifdef DRAW
void world_draw(const char* prefix, unsigned suffix)
{
    MagickWand* m_wand = NULL;
    PixelWand* c_wand = NULL;
    DrawingWand* d_wand = NULL;

    char filename[256];

    MagickWandGenesis();

    m_wand = NewMagickWand();
    c_wand = NewPixelWand();
    d_wand = NewDrawingWand();
    PixelSetColor(c_wand, "white");
    MagickNewImage(m_wand, world.size.width, world.size.height, c_wand);

    unsigned x,y;
    Field* wf;
    List_element* el;
    Host* host;

    for (x = 0; x < world.field_matrix->m; ++x) {
        for (y = 0; y < world.field_matrix->n; ++y) {
            wf = get_field_from_matrix_for_coord(x, y);
            el = list_first(wf->healthy);
            while (el) {
                host = list_element_get_value(el);
                switch (host->state) {
                    case HEALTHY:
                      PixelSetColor(c_wand, "green");
                      break;
                    case ILL:
                      PixelSetColor(c_wand, "red");
                      fprintf(stderr, "ill host in healthy list!\n");
                      break;
                    case IMMUNE:
                      PixelSetColor(c_wand, "blue");
                      break;
                    case INFECTED:
                      PixelSetColor(c_wand, "yellow");
                      fprintf(stderr, "infected host in healthy list!\n");
                      break;
                    case DEAD:
                      PixelSetColor(c_wand, "black");
                      fprintf(stderr, "dead host in healthy list!\n");
                      break;
                    default:
                      PixelSetColor(c_wand, "cyan");
                      printf("warning: unknown host state!\n");

                }
                DrawSetFillColor(d_wand, c_wand);
                DrawPoint(d_wand, host->pos.x, host->pos.y);
                el = list_element_next(el);
            }
            el = list_first(wf->infected);
            while (el) {
                host = list_element_get_value(el);
                switch (host->state) {
                    case HEALTHY:
                      PixelSetColor(c_wand, "green");
                      fprintf(stderr, "healthy host in infected list!\n");
                      break;
                    case ILL:
                      PixelSetColor(c_wand, "red");
                      break;
                    case IMMUNE:
                      PixelSetColor(c_wand, "blue");
                      fprintf(stderr, "immune host in infected list!\n");
                      break;
                    case INFECTED:
                      PixelSetColor(c_wand, "yellow");
                      break;
                    case DEAD:
                      PixelSetColor(c_wand, "black");
                      break;
                    default:
                      PixelSetColor(c_wand, "cyan");
                      printf("warning: unknown host state!\n");

                }
                DrawSetFillColor(d_wand, c_wand);
                DrawPoint(d_wand, host->pos.x, host->pos.y);
                el = list_element_next(el);
            }
        }
    }

    memset(filename, '\0', 256);
    sprintf(filename, "%s%u.png", prefix, suffix);

    MagickDrawImage(m_wand, d_wand);

    MagickWriteImage(m_wand,filename);
    c_wand = DestroyPixelWand(c_wand);
    m_wand = DestroyMagickWand(m_wand);
    d_wand = DestroyDrawingWand(d_wand);

    MagickWandTerminus();
}
#endif

void world_simulate(unsigned short time_steps, unsigned stop_if_finished)
{
    unsigned short time = 0;
    #ifdef DRAW
        #ifdef ALEX
            world_draw("/Users/a3z47u/Documents/text/Aktuell/Coding/Xcode_files/dkrz_projects/Infektsimulation/PICT/ts_", 0);
        #endif
        #ifdef FLORIAN
            world_draw("~", 0);
        #endif
    #endif

    unsigned x,y;
    Field* wf;
    Host* host;
    List_element* el;

    printf("time;healthy;infected;ill;died;infectious_dead;sum;infected_total;immune_total;ill_total\n");
    printf("%u;%u;%u;%u;%u;%u;%u;%u;%u;%u\n", time, world.healthy_current, 
      world.infected_current, world.ill_current, world.died_total, 
      world.infectious_dead_current, 
      world.infected_current + world.ill_current + world.healthy_current + world.died_total,
      world.infected_total, world.immune_total, world.ill_total);
    unsigned action_left = world.infected_current + world.ill_current + world.infectious_dead_current;
    while (time < time_steps && (!stop_if_finished || action_left > 0)) {
        /* move all hosts */ 
       for (x = 0; x < world.field_matrix->m; ++x) {
            for (y = 0; y < world.field_matrix->n; ++y) {
                wf = get_field_from_matrix_for_coord(x, y);
                //move all healthy
                el = list_first(wf->healthy);
                while (el) {
                    host = list_element_get_value(el);
                    el = list_element_next(el);
                    host_move(host);
                }
                //move all infected
                el = list_first(wf->infected);
                while (el) {
                    host = list_element_get_value(el);
                    el = list_element_next(el);
                    if (host->state != DEAD) {
                        host_move(host);
                    }
                }
            }
        }       
        /* process all hosts */
        for (x = 0; x < world.field_matrix->m; ++x) {
            for (y = 0; y < world.field_matrix->n; ++y) {
                wf = get_field_from_matrix_for_coord(x, y);
                //process all healthy
                el = list_first(wf->healthy);
                while (el) {
                    host = list_element_get_value(el);
                    el = list_element_next(el);
                    host_get_infection(host, time);
                    host_process(host, time);
                }
                //process all infected
                el = list_first(wf->infected);
                while (el) {
                    host = list_element_get_value(el);
                    el = list_element_next(el);
                    /* muss vor host_process gerufen werden,
                      falls host gelöscht wird (DEAD->host_delete) */
                    host_process(host, time);
                }
            }
        }       
 
        time++;
        printf("%u;%u;%u;%u;%u;%u;%u;%u;%u;%u\n", time, world.healthy_current, 
          world.infected_current, world.ill_current, world.died_total, 
          world.infectious_dead_current, 
          world.infected_current + world.ill_current + world.healthy_current + world.died_total,
          world.infected_total, world.immune_total, world.ill_total);
        
        action_left = world.infected_current + world.ill_current + world.infectious_dead_current;
 
        #ifdef DRAW
            #ifdef ALEX
                world_draw("/Users/a3z47u/Documents/text/Aktuell/Coding/Xcode_files/dkrz_projects/Infektsimulation/PICT/ts_", time);
            #endif
            #ifdef FLORIAN
                world_draw("~", time);
            #endif
        #endif
    }
}
